﻿Framework sidesign.minib3d
Import brl.eventqueue
Import brl.timer

Graphics3D 1920, 1200, 32, 2

' setting variables
Const mapname:String = "maps/goldrush.bsp"          ' mapname.bsp
Local CamMode:Int = 2
Local CamZoom:Float = 1.0
Local Rotation:Int = 1
Local XrayColors:Int[] = [128, 144, 160]

' other Variables
Local MapSize:Float
Local RotationX:Float = 0
Local RotationZ:Float = 0
Local RotAddX:Float
Local RotAddZ:Float

' Flip Timer
Global FlipTimer:TTimer = CreateTimer(60)

' 3D Objects
Local Camera:TCamera
Local Lightsource:TLight
Local Pivot:TPivot

' Load the BSP file and assemble a new Mesh from it
Global bsp:TBSP = New TBSP
bsp.BSP_TextureExcludes = ["sky", "common", "skies", "haze", "siwa_water", "sun"]
bsp.LoadBSP(mapname)

' Central Pivot
pivot = CreatePivot()
PositionEntity pivot, 0, 0, 0

' Determine MapSize for Rotation
MapSize = Min(bsp.BSP_MapSizeX, bsp.BSP_MapSizeY) / 2.0

' Camera
Camera = CreateCamera(Pivot)

' directional light
Lightsource = CreateLight(1)
RotateEntity Lightsource, 45, 45, 45
AmbientLight 80, 96, 128

' add entities
AddEntityMarkers()



' --------------------------------------------------------------------------------
' Main loop
' --------------------------------------------------------------------------------
While Not AppTerminate()

	Local Muliplicator:Int = 1

	If KeyHit(KEY_ESCAPE) Then End

	' wait for timer
	Select WaitEvent()
	
		Case EVENT_TIMERTICK
		
		Select EventSource()
								
			Case FlipTimer
					
				' LMB = speedup movement
				If MouseDown(1) Then Muliplicator = 5
				
				' RMB = view mode
				If MouseHit(2) Then
				
					CamMode:+1
					CamMode:Mod 3
										
				EndIf
				
				' SPACE = control rotation
				If KeyHit(KEY_SPACE) Then Rotation = 1 - Rotation
				
				' KEYS 1-9 = distance
				If KeyHit(KEY_1) Then CamZoom = 1.0
				If KeyHit(KEY_2) Then CamZoom = 2.0
				If KeyHit(KEY_3) Then CamZoom = 3.0
				If KeyHit(KEY_4) Then CamZoom = 4.0
				If KeyHit(KEY_5) Then CamZoom = 5.0
				If KeyHit(KEY_6) Then CamZoom = 6.0
				If KeyHit(KEY_7) Then CamZoom = 7.0
				If KeyHit(KEY_8) Then CamZoom = 8.0
				If KeyHit(KEY_9) Then CamZoom = 9.0
					
				' calculate camera movement
				If Rotation Or MouseDown(1) Then
				
					RotAddX = 0.2 * Muliplicator
					RotAddZ = 0.2 * Muliplicator
					RotationX:-RotAddX
					RotationZ:+RotAddZ
					
				EndIf
				
				' move camera
				Local CamDist:Float = EntityDistance(Camera, pivot)
				CameraRange Camera, CamDist * 0.125, CamDist * 4
				CameraZoom Camera, CamZoom

				PositionEntity Camera, Sin(RotationX) * 1.5 * MapSize * bsp.BSP_MapsizeX * 1.0 / bsp.BSP_MapSizeY, MapSize, Cos(RotationZ) * 1.5 * MapSize * bsp.BSP_MapsizeY * 1.0 / bsp.BSP_MapSizeX
				PointEntity Camera, Pivot
												
				' Render Mode
				Select CamMode
			
					' colored only
					Case 0
			
						CameraClsColor Camera, 0, 0, 0
							
						ShowEntity bsp.BSP_Color
						HideEntity bsp.BSP_Model
						
						EntityAlpha bsp.BSP_Color, 1
						EntityAlpha bsp.BSP_Model, 1
						
						Wireframe 0
						RenderWorld
				
					' X-Ray wireframe colored
					Case 1
					
						CameraClsColor Camera, 0, 0, 0
			
						ShowEntity bsp.BSP_Color
						HideEntity bsp.BSP_Model
						
						Wireframe 1
						RenderWorld
						
					' X-Ray wireframe green
					Case 2
				
						CameraClsColor Camera, 0, 0, 0
				
						' X-Ray wireframe
						EntityColor bsp.BSP_Model, XrayColors[0], XrayColors[1], XrayColors[2]
						EntityBlend bsp.BSP_Model, 1
				
						HideEntity bsp.BSP_Color
						ShowEntity bsp.BSP_Model
						
						EntityAlpha bsp.BSP_Color, 0.1
						EntityAlpha bsp.BSP_Model, 0.1

						CameraClsMode Camera, 1, 0
						Wireframe 0
						RenderWorld
				
						HideEntity bsp.BSP_Color
						ShowEntity bsp.BSP_Model
						
						CameraClsMode Camera, 0, 1
						Wireframe 1
						RenderWorld
						
						CameraClsMode Camera, 1, 1
											
				End Select
								
				Flip True
			
		End Select
		
	End Select
	
Wend

End



' --------------------------------------------------------------------------------
' add special entities for strategic visualization
' --------------------------------------------------------------------------------
Function AddEntityMarkers()

	' add spawns and train splines
	Local temp:TMesh, flag:TMesh
	For Local e:TBSPentity = EachIn TBSPentity.List
		
		' Axis spawn
		If e.ClassName = "team_CTF_redspawn" Then
					
			temp = CreateCube(Pivot)
			ScaleEntity temp, 36, 72, 36
			EntityColor temp, 255, 0, 0
			EntityFX temp, 1
			PositionEntity temp, e.x + bsp.BSP_OffX, e.y, e.z + bsp.BSP_OffZ
			
		EndIf
		
		' Allied spawn
		If e.ClassName = "team_CTF_bluespawn" Then
				
			temp = CreateCube(pivot)
			ScaleEntity temp, 36, 72, 36
			EntityColor temp, 0, 0, 255
			EntityFX temp, 1
			PositionEntity temp, e.x + bsp.BSP_OffX, e.y, e.z + bsp.BSP_OffZ
			
		EndIf
		
		' Objective
		If e.ClassName = "team_WOLF_objective" Or e.ClassName = "team_CTF_blueflag" Or e.ClassName = "team_CTF_redflag" Or e.ClassName = "func_constructible" Then
				
			temp = CreateCube(pivot)
			ScaleEntity temp, 72, 72, 72
			EntityColor temp, 255, 0, 255
			EntityFX temp, 1
			PositionEntity temp, e.x + bsp.BSP_OffX, e.y, e.z + bsp.BSP_OffZ
			
		EndIf
		
		' capture the flag
		If e.ClassName = "team_WOLF_checkpoint" Then
					
			temp = CreateCube(pivot)
			ScaleEntity temp, 8, 256, 8
			EntityColor temp, 255, 0, 255
			EntityFX temp, 1
			
			flag = CreateCube()
			ScaleEntity flag, 8, 64, 96
			EntityColor flag, 255, 0, 255
			EntityFX flag, 1
			
			PositionEntity temp, e.x + bsp.BSP_OffX, e.y + 256, e.z + bsp.BSP_OffZ
			PositionEntity flag, e.x + bsp.BSP_OffX, e.y + 512 - 64, e.z + bsp.BSP_OffZ - 96
			
		EndIf
				
		' health/ammo cabinets
		If e.ClassName = "misc_cabinet_supply" Or e.ClassName = "misc_cabinet_health" Then
				
			temp = CreateCube(pivot)
			ScaleEntity temp, 36, 72, 36
			EntityColor temp, 255, 255, 0
			EntityFX temp, 1
			PositionEntity temp, e.x + bsp.BSP_OffX, e.y, e.z + bsp.BSP_OffZ
			
		EndIf
		
		' train splines
		If e.ClassName = "info_train_spline_main" Then
				
			temp = CreateCube(pivot)
			ScaleEntity temp, 8, 2, 8
			EntityColor temp, 255, 255, 0
			EntityFX temp, 1
			PositionEntity temp, e.x + bsp.BSP_OffX, e.y, e.z + bsp.BSP_OffZ
			
		EndIf
	
		' train splines control marker
		If e.ClassName = "info_train_spline_control" Then
				
			temp = CreateCube(Pivot)
			ScaleEntity temp, 8, 2, 8
			EntityColor temp, 255, 0, 255
			EntityFX temp, 1
			PositionEntity temp, e.x + bsp.BSP_OffX, e.y, e.z + bsp.BSP_OffZ
			
		EndIf
		
	Next	

End Function



' --------------------------------------------------------------------------------
' 2D Distance calculation
' --------------------------------------------------------------------------------
Function Distance2D:Float(x1:Float, y1:Float, x2:Float, y2:Float)
	
	Local x:Float=x1-x2
	Local y:Float=y1-y2
	
	Return Sqr(x * x + y * y)
	
End Function



' ----------------------------------------------------------------------------
' Normalizes a value to given range
' ----------------------------------------------------------------------------
Function Normalize:Float(value:Float = 128.0, value_min:Float = 0.0, value_max:Float = 255.0, norm_min:Float = 0.0, norm_max:Float = 1.0, limit:Int = False)

	' normalize	
	Local result:Float=((value-value_min)/(value_max-value_min))*(norm_max-norm_min)+norm_min

	' limit	
	If value>norm_max Then value=norm_max Else If value<norm_min Then value=norm_min

	Return result
	
End Function



' --------------------------------------------------------------------------------
' Returns the maximum value of an array
' --------------------------------------------------------------------------------
Function MinMax:Int(nums:Int[], order:Int = 0)

	Local tmp:Int[] = nums[..]
	tmp.Sort()
	
	If order = 0 Then Return tmp[0] Else Return tmp[tmp.length - 1]
	
End Function



' --------------------------------------------------------------------------------
' Simple digit formatting output of a number
' --------------------------------------------------------------------------------
Function DigOut:String(in:String, delimiter:String = ".")

	Local l:Int = Len(in)
	Local j:Int
	Local negative:Int
	Local out:String
	
	If Int(in) < 0 Then negative = 1
	
	For j = l To 0 Step - 1

		out:+Mid(in, l - j, 1)

		If (j Mod 3) = 0 And j > 0 And (j < l - negative) Then out:+"."

	Next
	
	Return out

End Function



' --------------------------------------------------------------------------------
' Simple leading spaces for nice number formatting output
' --------------------------------------------------------------------------------
Function FormOut:String(in:String, spaces:Int = 2, prefix:String = " ")

	Local i:Int
	Local l:Int = Len(in)
	Local out:String
		
	If l < spaces Then
	
		For i = 0 To spaces - l - 1
		
			out:+prefix
		
		Next
	
	End If
	
	Return out + in
	
End Function




' --------------------------------------------------------------------------------
' Simple spacing for nice string formatting output
' --------------------------------------------------------------------------------
Function FormStr:String(in:String, spaces:Int = 2, suffix:String = " ")

	Local i:Int
	Local l:Int = Len(in)
	Local out:String
		
	If l < spaces Then
	
		For i = 0 To spaces - l - 1
		
			out:+suffix
		
		Next
	
	End If
	
	Return in + out
	
End Function


' --------------------------------------------------------------------------------
' BSP Library
' --------------------------------------------------------------------------------
Type TBSP

	' Constants
	Const BSP_MAX_LUMPS:Int = 16
	Const BSP_MAX_TEXTURES:Int = 255 * 4
	Const BSP_MAX_VERTS:Int = 65535 * 4
	Const BSP_MAX_FACES:Int = 65535 * 4
	Const BSP_MAX_LMAPS:Int = 1023 * 4
	Const BSP_MAX_SURFS:Int = 255 * 4
	Const BSP_MAX_MODELS:Int = 1023 * 4

	' Map Data
	Field BSP_EntityCount:Int
	Field BSP_TextureCount:Int
	Field BSP_ModelCount:Int
	Field BSP_VertexCount:Int
	Field BSP_FaceCount:Int
	Field BSP_CalcTime:Int
	Field BSP_Header:String
	Field BSP_Version:Int
	Field BSP_TextureExcludes:String[]
	
	' Map boundaries #1
	Field BSP_MinX:Float = 0.0
	Field BSP_MinY:Float = 0.0
	Field BSP_MinZ:Float = 0.0
	Field BSP_MaxX:Float = 0.0
	Field BSP_MaxY:Float = 0.0
	Field BSP_MaxZ:Float = 0.0
	
	' Map boundaries #2
	Field BSP_X1:Int
	Field BSP_X2:Int
	Field BSP_Y1:Int
	Field BSP_Y2:Int
	
	' Mapsize
	Field BSP_MapSizeX:Int
	Field BSP_MapSizeY:Int
	
	' Offsets
	Field BSP_OffX:Int
	Field BSP_OffZ:Int
	
	' Meshes
	Global BSP_Model:TMesh
	Global BSP_Color:TMesh
	
	' BSP
	Global BSP_Split:String[99]
	Global BSP_LumpOffset:Int[BSP_MAX_LUMPS + 1]
	Global BSP_LumpLenght:Int[BSP_MAX_LUMPS + 1]
	
	' Texture
	Global BSP_TextureName:String[BSP_MAX_TEXTURES + 1]
	Global BSP_TextureFlags:Int[BSP_MAX_TEXTURES + 1]
	Global BSP_TextureContent:Int[BSP_MAX_TEXTURES + 1]
	Global BSP_TextureHandle:TTexture[BSP_MAX_TEXTURES + 1]
	
	' Vertex
	Global BSP_VertexCoordX:Float[BSP_MAX_VERTS + 1]
	Global BSP_VertexCoordY:Float[BSP_MAX_VERTS + 1]
	Global BSP_VertexCoordZ:Float[BSP_MAX_VERTS + 1]
	Global BSP_VertexTexCoordU:Float[BSP_MAX_VERTS + 1]
	Global BSP_VertexLightmapCoordU:Float[BSP_MAX_VERTS + 1]
	Global BSP_VertexTexCoordV:Float[BSP_MAX_VERTS + 1]
	Global BSP_VertexLightmapCoordV:Float[BSP_MAX_VERTS + 1]
	Global BSP_VertexNX:Float[BSP_MAX_VERTS + 1]
	Global BSP_VertexNY:Float[BSP_MAX_VERTS + 1]
	Global BSP_VertexNZ:Float[BSP_MAX_VERTS + 1]
	Global BSP_VertexRed:Int[BSP_MAX_VERTS + 1]
	Global BSP_VertexGreen:Int[BSP_MAX_VERTS + 1]
	Global BSP_VertexBlue:Int[BSP_MAX_VERTS + 1]
	Global BSP_VertexAlpha:Float[BSP_MAX_VERTS + 1]
	
	' Mesh
	Global BSP_MeshVertexOffset:Int[BSP_MAX_VERTS + 1]
	
	' Face
	Global BSP_FaceTexture:Int[BSP_MAX_FACES + 1]
	Global BSP_FaceVertex:Int[BSP_MAX_FACES + 1]
	Global BSP_FaceNormal:Int[BSP_MAX_FACES + 1]
	Global BSP_FaceMeshVertex:Int[BSP_MAX_FACES + 1]
	Global BSP_FaceMeshNumber:Int[BSP_MAX_FACES + 1]
	Global BSP_FaceType:Int[BSP_MAX_FACES + 1]
	Global BSP_FaceLightmap:Int[BSP_MAX_FACES + 1]
	Global BSP_FaceNormalmap:Int[BSP_MAX_FACES + 1]
	Global BSP_FacePatchDimsX:Int[BSP_MAX_FACES + 1]
	Global BSP_FacePatchDimsY:Int[BSP_MAX_FACES + 1]
	
	' Model
	Global BSP_ModelBoxXstart:Float[BSP_MAX_MODELS + 1]
	Global BSP_ModelBoxYstart:Float[BSP_MAX_MODELS + 1]
	Global BSP_ModelBoxZstart:Float[BSP_MAX_MODELS + 1]
	Global BSP_ModelBoxXend:Float[BSP_MAX_MODELS + 1]
	Global BSP_ModelBoxYend:Float[BSP_MAX_MODELS + 1]
	Global BSP_ModelBoxZend:Float[BSP_MAX_MODELS + 1]
	Global BSP_ModelCenterX:Float[BSP_MAX_MODELS + 1]
	Global BSP_ModelCenterY:Float[BSP_MAX_MODELS + 1]
	Global BSP_ModelCenterZ:Float[BSP_MAX_MODELS + 1]
	Global BSP_ModelScaleX:Float[BSP_MAX_MODELS + 1]
	Global BSP_ModelScaleY:Float[BSP_MAX_MODELS + 1]
	Global BSP_ModelScaleZ:Float[BSP_MAX_MODELS + 1]
	Global BSP_ModelFacestart:Int[BSP_MAX_MODELS + 1]
	Global BSP_ModelFaceend:Int[BSP_MAX_MODELS + 1]
	
	' Lightmap
	Global BSP_LightmapHandle:TTexture[BSP_MAX_LMAPS, 2]
	
	' Surface
	Global BSP_SurfaceCount:Int[BSP_MAX_MODELS + 1]
	Global BSP_SurfaceTexture:Int[BSP_MAX_SURFS + 1]
	Global BSP_SurfaceLightmap:Int[BSP_MAX_SURFS + 1]
	Global BSP_SurfaceNormalmap:Int[BSP_MAX_SURFS + 1]

	
	
	' --------------------------------------------------------------------------------
	' BSP Main loader
	' --------------------------------------------------------------------------------
	Method LoadBSP(mapname:String)
	
		Local i:Int
		Local stream:TStream
		Local bank:TBank
		Local TotalCalcTime:Int = MilliSecs()
		Local header:String

		BSP_CalcTime = MilliSecs()
		
		Print "Parsing " + mapname

		stream = ReadFile(mapname)
		'If Not stream Then RuntimeError "Couldn't find BSP file!"
		BSP_Header = ReadChar(stream, 4)
		BSP_Version = ReadInt(stream)
				
		'If Not BSP_Header = "IBSP" Then RuntimeError "This is not a valid Q3BSP file!"
		
		Select BSP_Version
			
			Case $2E header = "Quake 3"
			Case $2F header = "RTCW/ET"
			
			Default header = "Unknown"
			
		End Select
		
		Print
		Print FormStr("Header:", 16) + FormOut(header, 8)

		For i = 0 To BSP_MAX_LUMPS - 1

			BSP_LumpOffset[i] = ReadInt(stream)
			BSP_LumpLenght[i] = ReadInt(stream)

		Next

		Print FormStr("Lumps:", 16) + FormOut(DigOut(BSP_MAX_LUMPS), 8)
		
		stream.seek(BSP_LumpOffset[0])
		bank = CreateBank(BSP_LumpLenght[0])
		
		For i = 0 To BankSize(bank) - 3
		
			PokeByte(bank, i, ReadByte(stream))

		Next
				
		For i = 0 To BankSize(bank) - 3
			
			' search for "{"
			If PeekByte(bank, i) = $7B Then
						
				' search for a following CR and "?" -> Entity found!
				If PeekByte(bank, i + 1) = $0A And PeekByte(bank, i + 2) = $22 Then ParseEntity(bank, i + 2)

			EndIf
		
		Next
		
		bank = Null

		Print FormStr("Entities:", 16) + FormOut(DigOut(BSP_EntityCount), 8)

		stream.seek(BSP_LumpOffset[1])

		For i = 0 To (BSP_LumpLenght[1] - 72) / 72
		
			BSP_TextureName[i] = ReadChar(stream, 64)
			BSP_TextureFlags[i] = ReadInt(stream)
			BSP_TextureContent[i] = ReadInt(stream)
			LoadTextures()

			Local t:TBSPtexture = New TBSPtexture

			BSP_TextureCount = BSP_TextureCount + 1
			t.Name = BSP_TextureName[i]
					
		Next
		
		Print FormStr("Textures:", 16) + FormOut(DigOut(BSP_TextureCount), 8)

		stream.seek(BSP_LumpOffset[7])

		For i = 0 To (BSP_LumpLenght[7] - 40) / 40
		
			BSP_ModelBoxXstart[BSP_ModelCount] = ReadFloat(stream)
			BSP_ModelBoxZstart[BSP_ModelCount] = ReadFloat(stream)
			BSP_ModelBoxYstart[BSP_ModelCount] = ReadFloat(stream)
		
			BSP_ModelBoxXend[BSP_ModelCount] = ReadFloat(stream)
			BSP_ModelBoxZend[BSP_ModelCount] = ReadFloat(stream)
			BSP_ModelBoxYend[BSP_ModelCount] = ReadFloat(stream)
		
			BSP_ModelScaleX[BSP_ModelCount] = BSP_ModelBoxXend[BSP_ModelCount] - BSP_ModelBoxXstart[BSP_ModelCount]
			BSP_ModelScaleY[BSP_ModelCount] = BSP_ModelBoxYend[BSP_ModelCount] - BSP_ModelBoxYstart[BSP_ModelCount]
			BSP_ModelScaleZ[BSP_ModelCount] = BSP_ModelBoxZend[BSP_ModelCount] - BSP_ModelBoxZstart[BSP_ModelCount]
		
			BSP_ModelCenterX[BSP_ModelCount] = BSP_ModelBoxXstart[BSP_ModelCount] + (BSP_ModelScaleX[BSP_ModelCount] / 2.0)
			BSP_ModelCenterY[BSP_ModelCount] = BSP_ModelBoxYstart[BSP_ModelCount] + (BSP_ModelScaleY[BSP_ModelCount] / 2.0)
			BSP_ModelCenterZ[BSP_ModelCount] = BSP_ModelBoxZstart[BSP_ModelCount] + (BSP_ModelScaleZ[BSP_ModelCount] / 2.0)
		
			BSP_ModelFacestart[BSP_ModelCount] = ReadInt(stream)
			BSP_ModelFaceend[BSP_ModelCount] = BSP_ModelFacestart[BSP_ModelCount] + ReadInt(stream) - 1
		
			ReadInt(stream)
			ReadInt(stream)
		
			BSP_ModelCount = BSP_ModelCount + 1
		
		Next
		
		Print FormStr("Models:", 16) + FormOut(DigOut(BSP_ModelCount), 8)

		stream.seek(BSP_LumpOffset[10])

		For i = 0 To (BSP_LumpLenght[10] - 1) / 44
		
			BSP_VertexCoordX[i] = ReadFloat(stream)
			BSP_VertexCoordY[i] = ReadFloat(stream)
			BSP_VertexCoordZ[i] = ReadFloat(stream)
		
			BSP_VertexTexCoordU[i] = ReadFloat(stream)
			BSP_VertexTexCoordV[i] = ReadFloat(stream)
			BSP_VertexLightmapCoordU[i] = ReadFloat(stream)
			BSP_VertexLightmapCoordV[i] = ReadFloat(stream)
		
			BSP_VertexNX[i] = ReadFloat(stream)
			BSP_VertexNY[i] = ReadFloat(stream)
			BSP_VertexNZ[i] = ReadFloat(stream)
		
			BSP_VertexRed[i] = ReadByte(stream)
			BSP_VertexGreen[i] = ReadByte(stream)
			BSP_VertexBlue[i] = ReadByte(stream)
			BSP_VertexAlpha[i] = ReadByte(stream) / 255.0
		
			BSP_VertexCount = BSP_VertexCount + 1
			'If BSP_VertexCount > BSP_MAX_VERTS Then RuntimeError "Maximum Vertices reached! Stopping Program!"
		
		Next
		
		stream.seek(BSP_LumpOffset[11])
		
		For i = 0 To (BSP_LumpLenght[11] - 1) / 4

			BSP_MeshVertexOffset[i] = ReadInt(stream)

		Next
		
		Print FormStr("Vertices:", 16) + FormOut(DigOut(BSP_VertexCount), 8)

		stream.seek(BSP_LumpOffset[13])

		For i = 0 To (BSP_LumpLenght[13] - 104) / 104
		
			BSP_FaceTexture[i] = ReadInt(stream)
			ReadInt(stream)

			BSP_FaceType[i] = ReadInt(stream)
			BSP_FaceVertex[i] = ReadInt(stream)
			BSP_FaceNormal[i] = ReadInt(stream)
			BSP_FaceMeshVertex[i] = ReadInt(stream)
			BSP_FaceMeshNumber[i] = ReadInt(stream)
			BSP_FaceLightmap[i] = ReadInt(stream)
			BSP_FaceNormalmap[i] = BSP_FaceLightmap[i]

			If BSP_FaceLightmap[i] < 0 Then BSP_FaceLightmap[i] = 0
			If BSP_FaceNormalmap[i] < 0 Then BSP_FaceNormalmap[i] = 0

			stream.seek(StreamPos(stream) + 64)

			BSP_FacePatchDimsX[i] = ReadInt(stream)
			BSP_FacePatchDimsY[i] = ReadInt(stream)

			BSP_FaceCount = BSP_FaceCount + 1
		
		Next
		
		Print FormStr("Faces:", 16) + FormOut(DigOut(BSP_FaceCount), 8)

		BSP_CalcTime = MilliSecs() - BSP_CalcTime

		Print
		Print "Parsing done in " + BSP_CalcTime + "ms"

		BSP_CalcTime = MilliSecs()

		Print
		Print "Assembling Map:"

		Q3BSP_AssembleModel()
		BSP_CalcTime = MilliSecs() - BSP_CalcTime

		Print
		Print "Assembling done in " + BSP_CalcTime + "ms"
		Print
		Print "Map built in " + (MilliSecs() - TotalCalcTime) + "ms"
		Print

		CloseFile(stream)
	
	End Method



	' --------------------------------------------------------------------------------
	' assemble map
	' --------------------------------------------------------------------------------
	Method Q3BSP_AssembleModel()
	
		Local m:Int, i:Int, surf1:TSurface, surf2:TSurface, texname:String
		Local index:Int, indexb:Int, indexc:Int, indexd:Int
		Local V0:Int, V1:Int, V2:Int
		Local tex:Int
		Local tris:Int, vert:Int
		Local ex:Int = 0
		Local s:Int, v:Int
		Local r:Int, g:Int, b:Int
		Local nx:Float, ny:Float, nz:Float, d:Float
			
		BSP_Model = CreateMesh(pivot)
		BSP_Color = CreateMesh(pivot)
		
		EntityFX BSP_Model, 1 + 16
		EntityFX BSP_Color, 2 + 16
		
		surf1 = CreateSurface(BSP_Model)
		surf2 = CreateSurface(BSP_Color)
						
		For m = 0 To BSP_Modelcount
				
			For i = BSP_ModelFacestart[m] To BSP_ModelFaceend[m]
									
				Select BSP_FaceType[i]
									
					' Standard & Models
					Case 1, 3
							
							If CountVertices(surf1) > 65000 Then
								surf1 = CreateSurface(BSP_Model)
								surf2 = CreateSurface(BSP_Color)
							EndIf
														
							texname = BSP_TextureName[BSP_FaceTexture[i]]
											
							tex = True
							For e = 0 To BSP_TextureExcludes.length - 1
							
								If Instr(texname, BSP_TextureExcludes[e]) Then
								
									tex = False
									ex:+1
									
								EndIf
				
							Next
											
							If tex Then
							
								r = 128
								g = 128
								b = 128
											
								' terrains
								If Instr(texname, "terrain") Then r = 128 ; g = 96 ; b = 64
								If Instr(texname, "sand") Then r = 128 ; g = 96 ; b = 64
								If Instr(texname, "cliff") Then r = 128 ; g = 96 ; b = 64
								If Instr(texname, "moss") Then r = 128 ; g = 96 ; b = 64
								
								' liquids
								If Instr(texname, "water") Then r = 0 ; g = 64 ; b = 128
								If Instr(texname, "river") Then r = 0 ; g = 64 ; b = 128
								If Instr(texname, "sea") Then r = 0 ; g = 64; b = 128
								If Instr(texname, "ice") Then r = 0 ; g = 64; b = 128
								
								' plants
								If Instr(texname, "tree") Then r = 64 ; g = 128 ; b = 64
			
								' roofs
								If Instr(texname, "roof") Then r = 128 ; g = 0 ; b = 0
								
								' roads
								If Instr(texname, "road") Then r = 0 ; g = 0 ; b = 0
													
								index = BSP_FaceVertex[i]
									
								For e = BSP_FaceMeshVertex[i] To BSP_FaceMeshVertex[i] + BSP_FaceMeshNumber[i] - 1 Step 3
								
									indexb = index + BSP_MeshVertexOffset[e + 0]
									indexc = index + BSP_MeshVertexOffset[e + 1]
									indexd = index + BSP_MeshVertexOffset[e + 2]
														
									' find X boundaries
									If BSP_VertexCoordX(indexb) < BSP_MinX Then BSP_MinX = BSP_VertexCoordX(indexb)
									If BSP_VertexCoordX(indexc) < BSP_MinX Then BSP_MinX = BSP_VertexCoordX(indexc)
									If BSP_VertexCoordX(indexd) < BSP_MinX Then BSP_MinX = BSP_VertexCoordX(indexd)
									If BSP_VertexCoordX(indexc) > BSP_MaxX Then BSP_MaxX = BSP_VertexCoordX(indexc)
									If BSP_VertexCoordX(indexb) > BSP_MaxX Then BSP_MaxX = BSP_VertexCoordX(indexb)
									If BSP_VertexCoordX(indexd) > BSP_MaxX Then BSP_MaxX = BSP_VertexCoordX(indexd)
									
									' find Y boundaries
									If BSP_VertexCoordY(indexb) < BSP_MinY Then BSP_MinY = BSP_VertexCoordY(indexb)
									If BSP_VertexCoordY(indexc) < BSP_MinY Then BSP_MinY = BSP_VertexCoordY(indexc)
									If BSP_VertexCoordY(indexd) < BSP_MinY Then BSP_MinY = BSP_VertexCoordY(indexd)
									If BSP_VertexCoordY(indexb) > BSP_MaxY Then BSP_MaxY = BSP_VertexCoordY(indexb)
									If BSP_VertexCoordY(indexc) > BSP_MaxY Then BSP_MaxY = BSP_VertexCoordY(indexc)
									If BSP_VertexCoordY(indexd) > BSP_MaxY Then BSP_MaxY = BSP_VertexCoordY(indexd)
									
									' find Z boundaries
									If BSP_VertexCoordZ(indexb) < BSP_MinZ Then BSP_MinZ = BSP_VertexCoordZ(indexb)
									If BSP_VertexCoordZ(indexc) < BSP_MinZ Then BSP_MinZ = BSP_VertexCoordZ(indexc)
									If BSP_VertexCoordZ(indexd) < BSP_MinZ Then BSP_MinZ = BSP_VertexCoordZ(indexd)
									If BSP_VertexCoordZ(indexb) > BSP_MaxZ Then BSP_MaxZ = BSP_VertexCoordZ(indexb)
									If BSP_VertexCoordZ(indexc) > BSP_MaxZ Then BSP_MaxZ = BSP_VertexCoordZ(indexc)
									If BSP_VertexCoordZ(indexd) > BSP_MaxZ Then BSP_MaxZ = BSP_VertexCoordZ(indexd)
														
									' vertices mesh
									V0 = AddVertex(surf1, BSP_VertexCoordX[indexb], BSP_VertexCoordZ[indexb], BSP_VertexCoordY[indexb] + mz, BSP_VertexTexCoordU[indexb], BSP_VertexTexCoordV[indexb])
									V1 = AddVertex(surf1, BSP_VertexCoordX[indexc], BSP_VertexCoordZ[indexc], BSP_VertexCoordY[indexc] + mz, BSP_VertexTexCoordU[indexc], BSP_VertexTexCoordV[indexc])
									V2 = AddVertex(surf1, BSP_VertexCoordX[indexd], BSP_VertexCoordZ[indexd], BSP_VertexCoordY[indexd] + mz, BSP_VertexTexCoordU[indexd], BSP_VertexTexCoordV[indexd])
												
									' vertex normals mesh
									VertexNormal surf1, V0, BSP_VertexNX[indexb], BSP_VertexNZ[indexb], BSP_VertexNY[indexb]
									VertexNormal surf1, V1, BSP_VertexNX[indexc], BSP_VertexNZ[indexc], BSP_VertexNY[indexc]
									VertexNormal surf1, V2, BSP_VertexNX[indexd], BSP_VertexNZ[indexd], BSP_VertexNY[indexd]
					
									' combine vertices to triangles
									AddTriangle(surf1, V0, V1, V2)
			
									' vertices mesh
									V0 = AddVertex(surf2, BSP_VertexCoordX[indexb] + sx, BSP_VertexCoordZ[indexb] + sy, BSP_VertexCoordY[indexb] + sz, BSP_VertexTexCoordU[indexb], BSP_VertexTexCoordV[indexb])
									V1 = AddVertex(surf2, BSP_VertexCoordX[indexc] + sx, BSP_VertexCoordZ[indexc] + sy, BSP_VertexCoordY[indexc] + sz, BSP_VertexTexCoordU[indexc], BSP_VertexTexCoordV[indexc])
									V2 = AddVertex(surf2, BSP_VertexCoordX[indexd] + sx, BSP_VertexCoordZ[indexd] + sy, BSP_VertexCoordY[indexd] + sz, BSP_VertexTexCoordU[indexd], BSP_VertexTexCoordV[indexd])
									
									' vertexx colors
									VertexColor surf2, V0, r, g, b
									VertexColor surf2, V1, r, g, b
									VertexColor surf2, V2, r, g, b
							
									' vertex normals mesh
									VertexNormal surf2, V0, BSP_VertexNX[indexb], BSP_VertexNZ[indexb], BSP_VertexNY[indexb]
									VertexNormal surf2, V1, BSP_VertexNX[indexc], BSP_VertexNZ[indexc], BSP_VertexNY[indexc]
									VertexNormal surf2, V2, BSP_VertexNX[indexd], BSP_VertexNZ[indexd], BSP_VertexNY[indexd]
					
									' combine vertices to triangles
									AddTriangle(surf2, V0, V1, V2)
															
								Next
						
							EndIf
							
					' Bezier Patches
					Case 2
							' tdb
			
				End Select
								
			Next
					
		Next
				
		' recalculate (normalize) normals
		For s = 1 To CountSurfaces(BSP_Model)
			
			surf1 = GetSurface(BSP_Model, s)
			surf2 = GetSurface(BSP_Color, s)
			
			For v = 0 To CountVertices(surf1) - 1
				
				nx = VertexNX(surf1, v)
				ny = VertexNY(surf1, v)
				nz = VertexNZ(surf1, v)
				d = 1.0 / Sqr(nx * nx + ny * ny + nz * nz)
				
				VertexNormal surf1, v, nx * d, ny * d, nz * d
				VertexNormal surf2, v, nx * d, ny * d, nz * d
				
			Next
			
			tris:+CountTriangles(surf1)
			vert:+CountVertices(surf1)
			
		Next
								
		' determine level boundaries
		BSP_X1 = Int(BSP_MinX)
		BSP_X2 = Int(BSP_MaxX)
		BSP_Y1 = Int(BSP_MinY)
		BSP_Y2 = Int(BSP_MaxY)
				
		' calculate mapsize	
		BSP_MapsizeX = (Abs(BSP_X1) + Abs(BSP_X2))
		BSP_MapsizeY = (Abs(BSP_Y1) + Abs(BSP_Y2))
		
		' calculate offset	
		BSP_OffX = (BSP_MapsizeX / 2) - BSP_X2
		BSP_OffZ = (BSP_MapsizeY / 2) - BSP_Y2
								
		' reposition meshes		
		PositionEntity BSP_Model, BSP_OffX, 0, BSP_OffZ
		PositionEntity BSP_Color, BSP_OffX, 0, BSP_OffZ
						
		Print FormStr("Surfaces:", 14) + FormOut(DigOut(CountSurfaces(BSP_Model)), 12)
		Print FormStr("Triangles:", 14) + FormOut(DigOut(tris), 12)
		Print FormStr("Vertices:", 14) + FormOut(DigOut(vert), 12)
		Print FormStr("Exclusions:", 14) + FormOut(DigOut(ex), 12)
		Print FormStr("Dimensions X:", 14) + FormOut(BSP_X1 + "," + BSP_X2, 12)
		Print FormStr("Dimensions Y:", 14) + FormOut(BSP_Y1 + "," + BSP_Y2, 12)
						
	End Method
	

	
	' --------------------------------------------------------------------------------
	' add missing zeros to a string
	' --------------------------------------------------------------------------------
	Method BSP_ZERO:String(number:String, lenght:Int = 4)
	
		Local i:Int
		Local r:String
	
		For i = 0 To lenght - Len(number) - 1

			r = r + "0"

		Next

		Return r + number
	
	End Method


	' --------------------------------------------------------------------------------
	' reads the textures from the BSP
	' --------------------------------------------------------------------------------
	Method LoadTextures()
	
		Local i:Int

		For i = 0 To BSP_MAX_TEXTURES
		
			BSP_TextureName[i] = Mid(BSP_TextureName[i], 1, Instr(BSP_TextureName[i], "*") - 1)

		Next

	End Method


	' --------------------------------------------------------------------------------
	' Reads a single character
	' --------------------------------------------------------------------------------
	Method ReadChar:String(stream:TStream, lenght:Int)
	
		Local char:String
		Local c:Int
		Local i:Int
	
		For i = 1 To lenght

			c = ReadByte(stream)
		
			If c = 0 Then

				char = char + "*"

			Else If c > 31 Then

				char = char + Chr(c)

			EndIf

		Next

		Return char

	End Method


	' --------------------------------------------------------------------------------
	' Parses a single BSP Entity
	' --------------------------------------------------------------------------------
	Method ParseEntity(bsp:TBank, index:Int)

		Local content:String
	
		' Create a new entity and store index as start position
		Local e:TBSPentity = New TBSPentity
	
		' as Long as there is no "}"
		While PeekByte(bsp, index) <> $7D
		
			' add bytes To String
			content = content + Chr(PeekByte(bsp, index))
			index = index + 1
		
		Wend
	
		' shorten string, replace " " with "=", replace spaces with commas, replace " and ; with nothing
		content = Mid(content, 1, Len(content) - 2)
		content = Replace(content, Chr(34) + Chr(32) + Chr(34), "=")
		content = Replace(content, Chr(32), "*")
		content = Replace(content, Chr(34), "")
		content = Replace(content, ";", "")
	
		' parse values of the entity
		GetEntityValues(e, content)

	End Method


	' --------------------------------------------------------------------------------
	' Patches a string's special characters
	' --------------------------------------------------------------------------------
	Method PatchUTF:String(t:String)
	
		t = Replace(t, Chr(195) + Chr(164), "ä")
		t = Replace(t, Chr(195) + Chr(132), "Ä")
		t = Replace(t, Chr(195) + Chr(182), "ö")
		t = Replace(t, Chr(195) + Chr(150), "ö")
		t = Replace(t, Chr(195) + Chr(188), "ü")
		t = Replace(t, Chr(195) + Chr(156), "Ü")
		t = Replace(t, Chr(195) + Chr(159), "ß")
	
		Return t
	
	End Method


	' --------------------------------------------------------------------------------
	' parses a single value
	' --------------------------------------------------------------------------------
	Method GetSingleValue:String(value:String, content:String)
	
		If Instr(content, value) Then Return SearchBetween(content, value, Chr(10))
	
	End Method


	' --------------------------------------------------------------------------------
	' returns text between two strings
	' --------------------------------------------------------------------------------
	Method SearchBetween:String(txt:String, txtbefore:String, txtbehind:String, fl:Int = 1)
	
		Local F:Int
		Local l:Int
		Local r:String
		
		F = Instr(txt, txtbefore, fl) + Len(txtbefore)
		l = Instr(txt, txtbehind, F)
		r = Mid(txt, F, l - F)
	
		Return r
	
	End Method


	' --------------------------------------------------------------------------------
	' returns text
	' --------------------------------------------------------------------------------
	Method GetSingleText:String(value:String, content:String)
	
		content = Replace(content, "*", " ")
		content = Replace(content, "'", Chr(34))

		If Instr(content, value) Then Return Trim(SearchBetween(content, value, Chr(10)))
	
	End Method


	' --------------------------------------------------------------------------------
	' rounds a value up or down
	' --------------------------------------------------------------------------------
	Method RoundValue:Float(x:Float)
	
		If x >= 0 Then Return Floor(x) Else Return Ceil(x)
	
	End Method


	' --------------------------------------------------------------------------------
	' parses all entity values
	' --------------------------------------------------------------------------------
	Method GetEntityValues(e:TBSPentity, content:String)
	
		e.angle = False
	
		content = PatchUTF(content)
	
		e.ClassName = GetSingleValue("classname=", content)
		e.Scale = GetSingleValue("modelscale=", content).toFloat()
		e.model = GetSingleText("model=", content)
		e.targetname = GetSingleValue("targetname=", content)
		e.target = GetSingleValue("target=", content)
		e.spawnflags = GetSingleValue("spawnflags=", content).toInt()
		e.angle = GetSingleValue("angle=", content).toFloat() - 90 Mod 360.0
	
		BSP_EntityCount = BSP_EntityCount + 1
	
		If Instr(content, "rotation=") Then

			e.angle = GetSingleValue("rotation=", content).toFloat() - 90 Mod 360.0

		EndIf
	
		' parse origin
		If Instr(content, "origin=") Then
		
			' split positions
			SplitString(SearchBetween(content, "origin=", Chr(10)), "*")
		
			' store position To Type (switch values to match Blitz 3D space)
			e.x = Floor(BSP_Split[0].toFloat())
			e.y = Floor(BSP_Split[2].toFloat())
			e.z = Floor(BSP_Split[1].toFloat())
			
			If e.x = 0 And e.y = 0 And e.z = 0 Then e.Remove()
		
		EndIf
	
		' parse position
		If Instr(content, "position=") Then
		
			' split positions
			SplitString(SearchBetween(content, "position=", Chr(10)), "*")
		
			' store position To Type (switch values To match Blitz 3D space)
			e.x = Floor(BSP_Split[0].toFloat())
			e.y = Floor(BSP_Split[2].toFloat())
			e.z = Floor(BSP_Split[1].toFloat())
		
		EndIf
	
		' parse color
		If Instr(content, "_color=") Then
		
			' split colors
			SplitString(SearchBetween(content, "_color=", Chr(10)), "*")
		
			' store RGB color values To Type
			e.r = Int(RoundValue(Float(BSP_Split[0]) * 255))
			e.g = Int(RoundValue(Float(BSP_Split[1]) * 255))
			e.b = Int(RoundValue(Float(BSP_Split[2]) * 255))
		
		EndIf
	
		' parse XYZ angles for models
		If Instr(content, "angles=") Then
		
			SplitString(SearchBetween(content, "angles=", Chr(10)), "*")
		
			e.pitch = (360 - Float(BSP_Split[2])) Mod 360.0
			e.yaw = Float(BSP_Split[1])
			e.roll = Float(BSP_Split[0]) * -1
		
		EndIf
	
	End Method

	' --------------------------------------------------------------------------------
	' splits a string using a delimiter into a blitzarray
	' --------------------------------------------------------------------------------
	Method SplitString:Int(txt:String, delimiter:String)
	
		Local P:Int
		Local i:Int
		Local part:String
	
		BSP_Split[P] = ""
	
		For i = 1 To Len(txt)
		
			part = Mid:String(txt, i, 1)
			If part = delimiter:String Then
				P = P + 1
				BSP_Split[P] = ""
			Else
				BSP_Split[P] = BSP_Split[P] + part
			EndIf
		
		Next
	
		Return P
	
	End Method

End Type



' --------------------------------------------------------------------------------
' Library Extension for Textures
' --------------------------------------------------------------------------------
Type TBSPTexture Extends TBSP

	Global List:TList = CreateList()

	Method New()

		ListAddLast(List, Self)

	End Method

	Field id:TTexture
	Field nm:TTexture
	Field Name:String

End Type



' --------------------------------------------------------------------------------
' Library Extension for Surfaces
' --------------------------------------------------------------------------------
Type TBSPSurface Extends TBSP
	
	Field id:TSurface
	Field texture:String
	
End Type



' --------------------------------------------------------------------------------
' Library Extension for Entities
' --------------------------------------------------------------------------------
Type TBSPentity Extends TBSP

	Global List:TList = CreateList()

	Method New()

		ListAddLast(List, Self)

	End Method
	
	Method Remove()

		ListRemove(List, Self)

	End Method

	Field ClassName:String
	Field target:String
	Field targetname:String
	Field spawnflags:Int
	Field model:String
	Field Scale:Float
	Field angle:Float
	Field pitch:Float
	Field yaw:Float
	Field roll:Float
	Field x:Float
	Field y:Float
	Field z:Float
	Field r:Int
	Field g:Int
	Field b:Int

End Type